/**********************************************************************
Copyright (c) 2012 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
   ...
**********************************************************************/
package org.datanucleus.store.neo4j;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.FetchPlan;
import org.datanucleus.PropertyNames;
import org.datanucleus.exceptions.NucleusException;
import org.datanucleus.identity.IdentityUtils;
import org.datanucleus.identity.OID;
import org.datanucleus.identity.OIDFactory;
import org.datanucleus.identity.SCOID;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.DiscriminatorMetaData;
import org.datanucleus.metadata.DiscriminatorStrategy;
import org.datanucleus.metadata.EmbeddedMetaData;
import org.datanucleus.metadata.FieldRole;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.MetaDataUtils;
import org.datanucleus.metadata.VersionMetaData;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.FieldValues;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.StoreManager;
import org.datanucleus.store.connection.ManagedConnection;
import org.datanucleus.store.neo4j.fieldmanager.FetchFieldManager;
import org.datanucleus.store.schema.naming.ColumnType;
import org.datanucleus.store.types.converters.TypeConverter;
import org.datanucleus.util.NucleusLogger;
import org.neo4j.cypher.javacompat.ExecutionEngine;
import org.neo4j.cypher.javacompat.ExecutionResult;
import org.neo4j.graphdb.GraphDatabaseService;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.index.Index;
import org.neo4j.graphdb.index.IndexHits;

/**
 * Series of convenience methods to aid in the conversion between POJO and Neo4j Node.
 */
public class Neo4jUtils
{
    /**
     * Method to return the Node representing the specified ObjectProvider.
     * @param graphDB Database service
     * @param op The object provider
     * @return The Node for this ObjectProvider (or null if not found)
     * @throws NucleusException if more than 1 Node is found matching this ObjectProvider!
     */
    public static Node getNodeForObjectProvider(GraphDatabaseService graphDB, ObjectProvider op)
    {
        AbstractClassMetaData cmd = op.getClassMetaData();
        ExecutionContext ec = op.getExecutionContext();
        return getNodeForObjectId(graphDB, ec, cmd, op.getInternalObjectId());
    }

    /**
     * Method to return the Node representing the object with the specified identity.
     * @param graphDB Database service
     * @param ec Execution Context
     * @param cmd Metadata for the root class that this could be an instance of
     * @param id The identity of the object
     * @return The Node for this object (or null if not found)
     * @throws NucleusException if more than 1 Node is found matching this object identity!
     */
    public static Node getNodeForObjectId(GraphDatabaseService graphDB, ExecutionContext ec, 
            AbstractClassMetaData cmd, Object id)
    {
        StoreManager storeMgr = ec.getStoreManager();

        if (cmd.pkIsDatastoreAttributed(storeMgr))
        {
            // Using Neo4j "node id" as the identity, so use getNodeById
            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                OID oid = (OID)id;
                Long key = (Long)oid.getKeyValue();
                if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
                {
                    NucleusLogger.DATASTORE_NATIVE.debug("Retrieving Node for id=" + key);
                }
                return graphDB.getNodeById(key);
            }
            else if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                int[] pkFieldNumbers = cmd.getPKMemberPositions();
                if (pkFieldNumbers.length == 1)
                {
                    Long key = (Long)ec.getApiAdapter().getTargetKeyForSingleFieldIdentity(id);
                    if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
                    {
                        NucleusLogger.DATASTORE_NATIVE.debug("Retrieving Node for id=" + key);
                    }
                    return graphDB.getNodeById(key);
                }
                else
                {
                    // TODO Composite id using identity on one field?
                }
            }
        }

        if (!graphDB.index().existsForNodes(Neo4jStoreManager.NODE_TYPE_INDEX_NAME))
        {
            // No index yet so the Node can't exist
            return null;
        }

        StringBuffer cypherString = new StringBuffer();

        // Start from the nodes of the required type
        cypherString.append("START n=node:" + Neo4jStoreManager.NODE_TYPE_INDEX_NAME + 
            "(" + Neo4jStoreManager.NODE_TYPE_INDEX_KEY + "=\"" + cmd.getFullClassName() + "\")");

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            // Application id - Add PK field(s) to the query object
            cypherString.append(" WHERE (");
            int[] pkPositions = cmd.getPKMemberPositions();
            for (int i=0;i<pkPositions.length;i++)
            {
                AbstractMemberMetaData pkMmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(pkPositions[i]);
                Object value = null;
                if (cmd.usesSingleFieldIdentityClass())
                {
                    value = ec.getApiAdapter().getTargetKeyForSingleFieldIdentity(id);
                }
                else
                {
                    value = IdentityUtils.getValueForMemberInId(id, pkMmd);
                }

                cypherString.append("n." + ec.getStoreManager().getNamingFactory().getColumnName(pkMmd, ColumnType.COLUMN));
                cypherString.append(" = ");
                Object storedValue = Neo4jUtils.getStoredValueForField(ec, pkMmd, value, FieldRole.ROLE_FIELD);
                if (storedValue instanceof String)
                {
                    cypherString.append("\"" + storedValue + "\"");
                }
                else
                {
                    cypherString.append(storedValue);
                }
                if (i != pkPositions.length-1)
                {
                    cypherString.append(" and ");
                }
            }
            cypherString.append(")");
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            OID oid = (OID)id;
            if (oid == null && storeMgr.isStrategyDatastoreAttributed(cmd, -1))
            {
                // Not yet set, so return null (needs to be attributed in the datastore)
                return null;
            }
            Object value = oid.getKeyValue();
            String propName = ec.getStoreManager().getNamingFactory().getColumnName(cmd, ColumnType.DATASTOREID_COLUMN);
            cypherString.append(" WHERE (n.");
            cypherString.append(propName);
            cypherString.append(" = ");
            cypherString.append(Neo4jUtils.getStoredValueForField(ec, null, value, FieldRole.ROLE_FIELD));
            cypherString.append(")");
        }
        else
        {
            throw new NucleusException("Impossible to retrieve Node for nondurable identity");
        }

        if (cmd.hasDiscriminatorStrategy())
        {
            DiscriminatorMetaData discmd = cmd.getDiscriminatorMetaData();
            String propName = storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DISCRIMINATOR_COLUMN);
            Object discVal = null;
            if (cmd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
            {
                discVal = cmd.getFullClassName();
            }
            else
            {
                discVal = discmd.getValue();
            }
            cypherString.append(" and (n.").append(propName).append(" = \"").append(discVal).append("\")");
        }

        cypherString.append(" RETURN n");

        if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_NATIVE.debug("Retrieving objects using Cypher query : " + cypherString);
        }

        ExecutionEngine engine = new ExecutionEngine(graphDB);
        ExecutionResult result = engine.execute(cypherString.toString());
        if (ec.getStatistics() != null)
        {
            // Add to statistics
            ec.getStatistics().incrementNumReads();
        }

        Iterator<Map<String, Object>> iter = result.iterator();
        if (!iter.hasNext())
        {
            return null;
        }

        Map<String, Object> map = iter.next();
        Node node = (Node)map.get("n");
        if (iter.hasNext())
        {
            throw new NucleusException("Query of Node for object with id=" + id + 
                " returned more than 1 result! : " + cypherString);
        }
        return node;
    }

    /**
     * Convenience method to return all objects of the candidate type (optionally allowing subclasses).
     * @param ec Execution context
     * @param db Neo4j connection
     * @param candidateClass Candidate
     * @param subclasses Whether to include subclasses
     * @param ignoreCache Ignore the cache?
     * @param fp FetchPlan when retrieving the objects
     * @param options Set of options for controlling this query
     * @param candidateAliasName Alias to use in Cypher query
     * @param filterText Text to use in the filter for the Cypher query (if any)
     * @param orderText Text to use in the order for the Cypher query (if any)
     * @return List of all candidate objects
     */
    public static List getObjectsOfCandidateType(ExecutionContext ec, GraphDatabaseService db, Class candidateClass, 
            boolean subclasses, boolean ignoreCache, FetchPlan fp, Map<String, Object> options,
            String candidateAliasName, String filterText, String orderText)
    {
        if (candidateAliasName == null)
        {
            candidateAliasName = "n";
        }
        AbstractClassMetaData cmd =
            ec.getMetaDataManager().getMetaDataForClass(candidateClass, ec.getClassLoaderResolver());
        StoreManager storeMgr = ec.getStoreManager();

        StringBuffer cypherString = new StringBuffer();

        // Start from the nodes of the required type
        cypherString.append("START " + candidateAliasName + "=node:" + Neo4jStoreManager.NODE_TYPE_INDEX_NAME + 
            "(" + Neo4jStoreManager.NODE_TYPE_INDEX_KEY + "=\"" + cmd.getFullClassName() + "\")");

        // Add any WHERE clause
        boolean multiple = false;
        String discriminatorText = null;
        if (cmd.hasDiscriminatorStrategy())
        {
            // Restriction on discriminator for this class
            DiscriminatorMetaData discmd = cmd.getDiscriminatorMetaData();
            String propName = storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DISCRIMINATOR_COLUMN);
            Object discVal = null;
            if (cmd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
            {
                discVal = cmd.getFullClassName();
            }
            else
            {
                discVal = discmd.getValue();
            }
            discriminatorText = propName + " = \"" + discVal + "\"";
            if (filterText != null)
            {
                multiple = true;
            }
        }
        String multitenancyText = null;
        if (storeMgr.getStringProperty(PropertyNames.PROPERTY_TENANT_ID) != null)
        {
            // Restriction on multitenancy discriminator for this tenant
            if ("true".equalsIgnoreCase(cmd.getValueForExtension("multitenancy-disable")))
            {
                // Don't bother with multitenancy for this class
            }
            else
            {
                String propName = storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.MULTITENANCY_COLUMN);
                String value = storeMgr.getStringProperty(PropertyNames.PROPERTY_TENANT_ID);
                multitenancyText = propName + " = \"" + value + "\"";
                if (filterText != null || discriminatorText != null)
                {
                    multiple = true;
                }
            }
        }
        if (filterText != null || discriminatorText != null || multitenancyText != null)
        {
            cypherString.append(" WHERE ");
            boolean started = false;
            if (filterText != null)
            {
                if (multiple)
                {
                    cypherString.append("(");
                }
                cypherString.append(filterText);
                if (multiple)
                {
                    cypherString.append(")");
                }
                started = true;
            }
            if (discriminatorText != null)
            {
                if (started)
                {
                    cypherString.append(" and");
                }
                if (multiple)
                {
                    cypherString.append("(");
                }
                cypherString.append(discriminatorText);
                if (multiple)
                {
                    cypherString.append(")");
                }
                started = true;
            }
            if (multitenancyText != null)
            {
                if (started)
                {
                    cypherString.append(" and");
                }
                if (multiple)
                {
                    cypherString.append("(");
                }
                cypherString.append(multitenancyText);
                if (multiple)
                {
                    cypherString.append(")");
                }
            }
        }

        cypherString.append(" RETURN " + candidateAliasName);
        if (orderText != null)
        {
            cypherString.append(" ORDER BY " + orderText);
        }

        if (NucleusLogger.DATASTORE_NATIVE.isDebugEnabled())
        {
            NucleusLogger.DATASTORE_NATIVE.debug("Retrieving objects using Cypher query : " + cypherString);
        }

        List results = new ArrayList();
        ExecutionEngine engine = new ExecutionEngine(db);
        ExecutionResult queryResult = engine.execute(cypherString.toString());
        Iterator<Map<String, Object>> iter = queryResult.iterator();
        fp.manageFetchPlanForClass(cmd);
        int[] fpMembers = fp.getFetchPlanForClass(cmd).getMemberNumbers();

        while (iter.hasNext())
        {
            Map<String, Object> map = iter.next();
            if (map.size() > 0)
            {
                Node node = (Node) map.get(candidateAliasName);
                AbstractClassMetaData nodeCmd = getClassMetaDataForNode(node, ec, cmd);
                if (!subclasses && !cmd.getFullClassName().equals(nodeCmd.getFullClassName()))
                {
                    // Is actually of a subtype and we aren't considering those
                    continue;
                }

                Object obj = getObjectForNode(node, nodeCmd, ec, ignoreCache, fpMembers);
                results.add(obj);
            }
        }

        return results;
    }

    /**
     * Convenience method to return the inheritance level of a Node, providing the root class that it
     * definitely is an instance of. Interrogates the node information for the types index to find the subclass
     * that this Node belongs to.
     * @param node The node
     * @param ec Execution Context
     * @param cmd Root metadata
     * @return AbstractClassMetaData for the type that this Node is an instance of.
     */
    public static AbstractClassMetaData getClassMetaDataForNode(Node node, ExecutionContext ec, AbstractClassMetaData cmd)
    {
        // Assumed to be of the root type, but check all subclasses
        Index<Node> typesIdx = node.getGraphDatabase().index().forNodes(Neo4jStoreManager.NODE_TYPE_INDEX_NAME);

        String[] subclassNames = ec.getMetaDataManager().getSubclassesForClass(cmd.getFullClassName(), false);
        if (subclassNames != null)
        {
            for (int i=0;i<subclassNames.length;i++)
            {
                AbstractClassMetaData subcmd = ec.getMetaDataManager().getMetaDataForClass(subclassNames[i], ec.getClassLoaderResolver());
                IndexHits<Node> nodesForSubclass = typesIdx.get(Neo4jStoreManager.NODE_TYPE_INDEX_KEY, subcmd.getFullClassName());
                if (nodesForSubclass != null)
                {
                    for (Node subclassNode : nodesForSubclass)
                    {
                        if (subclassNode.equals(node))
                        {
                            return getClassMetaDataForNode(subclassNode, ec, subcmd);
                        }
                    }
                }
            }
        }

        return cmd;
    }

    public static Object getObjectForNode(Node node, AbstractClassMetaData cmd, ExecutionContext ec,
            boolean ignoreCache, int[] fpMembers)
    {
        // TODO Process any discriminator property in the node to set "cmd"
        if (fpMembers == null)
        {
            ec.getFetchPlan().manageFetchPlanForClass(cmd);
            fpMembers = ec.getFetchPlan().getFetchPlanForClass(cmd).getMemberNumbers();
        }

        Object obj = null;
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            obj = getObjectUsingApplicationIdForDBObject(node, cmd, ec, ignoreCache, fpMembers);
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            obj = getObjectUsingDatastoreIdForDBObject(node, cmd, ec, ignoreCache, fpMembers);
        }
        else
        {
            obj = getObjectUsingNondurableIdForDBObject(node, cmd, ec, ignoreCache, fpMembers);
        }
        return obj;
    }

    public static Object getObjectUsingApplicationIdForDBObject(final Node node, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        Object id = IdentityUtils.getApplicationIdentityForResultSetRow(ec, cmd, null, 
            false, new FetchFieldManager(ec, node, cmd));

        StoreManager storeMgr = ec.getStoreManager();
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(id, 
            new FieldValues()
            {
                public void fetchFields(ObjectProvider op)
                {
                    op.replaceFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider op)
                {
                    op.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = node.getProperty(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }

    public static Object getObjectUsingDatastoreIdForDBObject(final Node node, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        StoreManager storeMgr = ec.getStoreManager();
        Object idKey = node.getProperty(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.DATASTOREID_COLUMN));

        OID oid = OIDFactory.getInstance(ec.getNucleusContext(), cmd.getFullClassName(), idKey);
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(oid, 
            new FieldValues()
            {
                // StateManager calls the fetchFields method
                public void fetchFields(ObjectProvider op)
                {
                    op.replaceFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider op)
                {
                    op.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = node.getProperty(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }
    public static Object getObjectUsingNondurableIdForDBObject(final Node node, final AbstractClassMetaData cmd,
            final ExecutionContext ec, boolean ignoreCache, final int[] fpMembers)
    {
        SCOID oid = new SCOID(cmd.getFullClassName());
        Class type = ec.getClassLoaderResolver().classForName(cmd.getFullClassName());
        Object pc = ec.findObject(oid, 
            new FieldValues()
            {
                // StateManager calls the fetchFields method
                public void fetchFields(ObjectProvider op)
                {
                    op.replaceFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public void fetchNonLoadedFields(ObjectProvider op)
                {
                    op.replaceNonLoadedFields(fpMembers, new FetchFieldManager(ec, node, cmd));
                }
                public FetchPlan getFetchPlanForLoading()
                {
                    return null;
                }
            }, type, ignoreCache, false);

        if (cmd.isVersioned())
        {
            // Set the version on the retrieved object
            ObjectProvider sm = ec.findObjectProvider(pc);
            StoreManager storeMgr = ec.getStoreManager();
            Object version = null;
            VersionMetaData vermd = cmd.getVersionMetaDataForClass();
            if (vermd.getFieldName() != null)
            {
                // Get the version from the field value
                AbstractMemberMetaData verMmd = cmd.getMetaDataForMember(vermd.getFieldName());
                version = sm.provideField(verMmd.getAbsoluteFieldNumber());
            }
            else
            {
                // Get the surrogate version from the datastore
                version = node.getProperty(storeMgr.getNamingFactory().getColumnName(cmd, ColumnType.VERSION_COLUMN));
            }
            sm.setVersion(version);
        }
        return pc;
    }

    /**
     * Convenience method to return the value to store for the specified field and value taking into account
     * the types that are supported in Neo4j.
     * @param ec ExecutionContext
     * @param mmd Metadata for the member
     * @param value The value of the field
     * @param fieldRole The role of the field
     * @return The value to store
     */
    public static Object getStoredValueForField(ExecutionContext ec, AbstractMemberMetaData mmd, Object value, int fieldRole)
    {
        if (value == null)
        {
            return null;
        }

        Class type = value.getClass();
        if (mmd != null)
        {
            if (fieldRole == FieldRole.ROLE_COLLECTION_ELEMENT)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getCollection().getElementType());
            }
            else if (fieldRole == FieldRole.ROLE_ARRAY_ELEMENT)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getArray().getElementType());
            }
            else if (fieldRole == FieldRole.ROLE_MAP_KEY)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getMap().getKeyType());
            }
            else if (fieldRole == FieldRole.ROLE_MAP_VALUE)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getMap().getValueType());
            }
            else
            {
                type = mmd.getType();
            }
        }

        if (Byte.class.isAssignableFrom(type) ||
            Boolean.class.isAssignableFrom(type) ||
            Character.class.isAssignableFrom(type) ||
            Double.class.isAssignableFrom(type) ||
            Float.class.isAssignableFrom(type) ||
            Integer.class.isAssignableFrom(type) ||
            Long.class.isAssignableFrom(type) ||
            Short.class.isAssignableFrom(type) ||
            String.class.isAssignableFrom(type))
        {
            // Natively supported
            return value;
        }
        else if (Enum.class.isAssignableFrom(type))
        {
            ColumnMetaData colmd = null;
            if (mmd != null && mmd.getColumnMetaData() != null && mmd.getColumnMetaData().length > 0)
            {
                colmd = mmd.getColumnMetaData()[0];
            }
            boolean useNumeric = MetaDataUtils.persistColumnAsNumeric(colmd);
            return useNumeric ? ((Enum)value).ordinal() : ((Enum)value).name();
        }

        // Fallback to built-in type converters
        TypeConverter strConv = ec.getTypeManager().getTypeConverterForType(type, String.class);
        TypeConverter longConv = ec.getTypeManager().getTypeConverterForType(type, Long.class);
        if (strConv != null)
        {
            // store as a String
            return strConv.toDatastoreType(value);
        }
        else if (longConv != null)
        {
            // store as a Long
            return longConv.toDatastoreType(value);
        }

        // TODO Cater for cases with no converters
        return value;
    }

    /**
     * Convenience method to convert the stored value for an object field into the value that will be held
     * in the object. Note that this does not cater for relation fields, just basic fields.
     * @param ec ExecutionContext
     * @param mmd Metadata for the field holding this value
     * @param value The stored value for the field
     * @param fieldRole The role of this value for the field
     * @return The value to put in the field
     */
    public static Object getFieldValueFromStored(ExecutionContext ec, AbstractMemberMetaData mmd, Object value, int fieldRole)
    {
        if (value == null)
        {
            return null;
        }

        Class type = value.getClass();
        if (mmd != null)
        {
            if (fieldRole == FieldRole.ROLE_COLLECTION_ELEMENT)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getCollection().getElementType());
            }
            else if (fieldRole == FieldRole.ROLE_ARRAY_ELEMENT)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getArray().getElementType());
            }
            else if (fieldRole == FieldRole.ROLE_MAP_KEY)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getMap().getKeyType());
            }
            else if (fieldRole == FieldRole.ROLE_MAP_VALUE)
            {
                type = ec.getClassLoaderResolver().classForName(mmd.getMap().getValueType());
            }
            else
            {
                type = mmd.getType();
            }
        }

        if (Byte.class.isAssignableFrom(type) ||
            Boolean.class.isAssignableFrom(type) ||
            Character.class.isAssignableFrom(type) ||
            Double.class.isAssignableFrom(type) ||
            Float.class.isAssignableFrom(type) ||
            Integer.class.isAssignableFrom(type) ||
            Long.class.isAssignableFrom(type) ||
            Short.class.isAssignableFrom(type) ||
            String.class.isAssignableFrom(type))
        {
            return value;
        }
        else if (Enum.class.isAssignableFrom(type))
        {
            ColumnMetaData colmd = null;
            if (mmd != null && mmd.getColumnMetaData() != null && mmd.getColumnMetaData().length > 0)
            {
                colmd = mmd.getColumnMetaData()[0];
            }
            if (MetaDataUtils.persistColumnAsNumeric(colmd))
            {
                return type.getEnumConstants()[((Number)value).intValue()];
            }
            else
            {
                return Enum.valueOf(type, (String)value);
            }
        }

        TypeConverter strConv = ec.getTypeManager().getTypeConverterForType(type, String.class);
        TypeConverter longConv = ec.getTypeManager().getTypeConverterForType(type, Long.class);
        if (strConv != null)
        {
            // Persisted as a String, so convert back
            String strValue = (String)value;
            return strConv.toMemberType(strValue);
        }
        else if (longConv != null)
        {
            // Persisted as a Long, so convert back
            Long longValue = (Long)value;
            return longConv.toMemberType(longValue);
        }

        // TODO Cater for cases with no converters
        return value;
    }

    /**
     * Convenience method to return the property (column) name for an embedded field.
     * @param ownerMmd Metadata for the field holding the embedded persistable object
     * @param fieldNumber FieldNumber of the embedded persistable object
     * @return Property name to use
     */
    public static String getPropertyNameForEmbeddedField(AbstractMemberMetaData ownerMmd, int fieldNumber)
    {
        String columnName = null;

        EmbeddedMetaData embmd = ownerMmd.getEmbeddedMetaData();
        AbstractMemberMetaData embMmd = null;
        if (embmd != null)
        {
            AbstractMemberMetaData[] embmmds = embmd.getMemberMetaData();
            embMmd = embmmds[fieldNumber];
        }

        ColumnMetaData[] colmds = embMmd.getColumnMetaData();
        if (colmds != null && colmds.length > 0)
        {
            // Try first column if specified
            columnName = colmds[0].getName();
        }
        if (columnName == null)
        {
            // Fallback to the field/property name
            columnName = embMmd.getName();
        }
        if (columnName == null)
        {
            columnName = embMmd.getName();
        }
        return columnName;
    }

    /**
     * Convenience method that returns the actual class that this identity is an instance of.
     * Makes use of the DN_TYPES index to find the ultimate subclass that it is an instanceof.
     * @param id The identity
     * @param rootCmd ClassMetaData for the root class in the inheritance tree
     * @param ec ExecutionContext
     * @param clr ClassLoader resolver
     * @return The class name of the object with this id
     */
    public static String getClassNameForIdentity(Object id, AbstractClassMetaData rootCmd, ExecutionContext ec,
            ClassLoaderResolver clr)
    {
        StoreManager storeMgr = ec.getStoreManager();
        ManagedConnection mconn = storeMgr.getConnection(ec);
        try
        {
            GraphDatabaseService db = (GraphDatabaseService)mconn.getConnection();
            Node node = Neo4jUtils.getNodeForObjectId(db, ec, rootCmd, id);
            AbstractClassMetaData cmd = getClassMetaDataForNode(node, ec, rootCmd);
            return cmd.getFullClassName();
        }
        finally
        {
            mconn.release();
        }
    }
}