/**********************************************************************
Copyright (c) 2012 Andy Jefferson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors:
    ...
**********************************************************************/
package org.datanucleus.store.neo4j.fieldmanager;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.exceptions.NucleusUserException;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.FieldRole;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.fieldmanager.AbstractStoreFieldManager;
import org.datanucleus.store.fieldmanager.FieldManager;
import org.datanucleus.store.neo4j.Neo4jStoreManager;
import org.datanucleus.store.neo4j.Neo4jUtils;
import org.datanucleus.store.schema.naming.ColumnType;
import org.datanucleus.store.types.TypeManager;
import org.datanucleus.store.types.converters.TypeConverter;
import org.neo4j.graphdb.Node;
import org.neo4j.graphdb.Relationship;

/**
 * Field Manager for putting values from a POJO into a Neo4j Node.
 */
public class StoreFieldManager extends AbstractStoreFieldManager
{
    /** Node that we are populating with properties representing the fields of the POJO. */
    protected Node node;

    /** Metadata of the owner field if this is for an embedded object. */
    protected AbstractMemberMetaData ownerMmd = null;

    public StoreFieldManager(ObjectProvider op, Node node, boolean insert)
    {
        super(op, insert);
        this.node = node;
    }

    protected String getPropName(int fieldNumber)
    {
        return op.getExecutionContext().getStoreManager().getNamingFactory().getColumnName(
            cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber), ColumnType.COLUMN);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeBooleanField(int, boolean)
     */
    @Override
    public void storeBooleanField(int fieldNumber, boolean value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeByteField(int, byte)
     */
    @Override
    public void storeByteField(int fieldNumber, byte value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeCharField(int, char)
     */
    @Override
    public void storeCharField(int fieldNumber, char value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeDoubleField(int, double)
     */
    @Override
    public void storeDoubleField(int fieldNumber, double value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeFloatField(int, float)
     */
    @Override
    public void storeFloatField(int fieldNumber, float value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeIntField(int, int)
     */
    @Override
    public void storeIntField(int fieldNumber, int value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeLongField(int, long)
     */
    @Override
    public void storeLongField(int fieldNumber, long value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeShortField(int, short)
     */
    @Override
    public void storeShortField(int fieldNumber, short value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeStringField(int, java.lang.String)
     */
    @Override
    public void storeStringField(int fieldNumber, String value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        if (value == null)
        {
            if (!insert)
            {
                node.removeProperty(getPropName(fieldNumber));
            }
            return;
        }
        node.setProperty(getPropName(fieldNumber), value);
    }

    /* (non-Javadoc)
     * @see org.datanucleus.store.fieldmanager.AbstractFieldManager#storeObjectField(int, java.lang.Object)
     */
    @Override
    public void storeObjectField(int fieldNumber, Object value)
    {
        AbstractMemberMetaData mmd = cmd.getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (!isStorable(mmd))
        {
            return;
        }

        ExecutionContext ec = op.getExecutionContext();
        String propName = ec.getStoreManager().getNamingFactory().getColumnName(mmd, ColumnType.COLUMN);
        if (!insert && node.hasProperty(propName) && value == null)
        {
            // Updating the field, it had a value but this time is null, so remove it
            node.removeProperty(propName);
            return;
        }

        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        int relationType = mmd.getRelationType(clr);

        boolean embedded = false;
        if (relationType != Relation.NONE)
        {
            // Determine if this relation field is embedded
            if (Relation.isRelationSingleValued(relationType))
            {
                if (ownerMmd != null && ownerMmd.getEmbeddedMetaData() != null)
                {
                    // Is this a nested embedded (from JDO definition) with specification for this field?
                    AbstractMemberMetaData[] embMmds = ownerMmd.getEmbeddedMetaData().getMemberMetaData();
                    if (embMmds != null)
                    {
                        for (int i=0;i<embMmds.length;i++)
                        {
                            if (embMmds[i].getName().equals(mmd.getName()))
                            {
                                embedded = true;
                            }
                        }
                    }
                }
            }

            if (mmd.isEmbedded())
            {
                // Is field marked as embedded?
                embedded = true;
            }
            else if (mmd.getEmbeddedMetaData() != null)
            {
                // Does this field have @Embedded definition?
                embedded = true;
            }
            else if (Relation.isRelationMultiValued(relationType))
            {
                // Is this an embedded element/key/value?
                if (mmd.hasCollection() && mmd.getElementMetaData() != null && mmd.getElementMetaData().getEmbeddedMetaData() != null)
                {
                    // Embedded collection element
                    embedded = true;
                }
                else if (mmd.hasArray() && mmd.getElementMetaData() != null && mmd.getElementMetaData().getEmbeddedMetaData() != null)
                {
                    // Embedded array element
                    embedded = true;
                }
                else if (mmd.hasMap() && 
                        ((mmd.getKeyMetaData() != null && mmd.getKeyMetaData().getEmbeddedMetaData() != null) || 
                        (mmd.getValueMetaData() != null && mmd.getValueMetaData().getEmbeddedMetaData() != null)))
                {
                    // Embedded map key/value
                    embedded = true;
                }
            }
        }

        if (embedded)
        {
            if (Relation.isRelationSingleValued(relationType) && value != null)
            {
                // Embedded PC object
                // TODO Cater for nulled embedded object on update
                if (ownerMmd != null)
                {
                    // Detect bidirectional relation so we know when to stop embedding
                    if (Relation.isBidirectional(relationType))
                    {
                        // Field has mapped-by, so just use that
                        if ((ownerMmd.getMappedBy() != null && mmd.getName().equals(ownerMmd.getMappedBy())) ||
                            (mmd.getMappedBy() != null && ownerMmd.getName().equals(mmd.getMappedBy())))
                        {
                            // This is other side of owner bidirectional, so omit
                            return;
                        }
                    }
                    else 
                    {
                        // mapped-by not set but could have owner-field
                        if (ownerMmd.getEmbeddedMetaData() != null &&
                            ownerMmd.getEmbeddedMetaData().getOwnerMember() != null &&
                            ownerMmd.getEmbeddedMetaData().getOwnerMember().equals(mmd.getName()))
                        {
                            // This is the owner-field linking back to the owning object so stop
                            return;
                        }
                    }
                }

                AbstractClassMetaData embcmd = ec.getMetaDataManager().getMetaDataForClass(value.getClass(), clr);
                if (embcmd == null)
                {
                    throw new NucleusUserException("Field " + mmd.getFullFieldName() +
                        " specified as embedded but metadata not found for the class of type " + mmd.getTypeName());
                }

                // Extract the owner member metadata for this embedded object
                AbstractMemberMetaData embMmd = mmd;
                if (ownerMmd != null)
                {
                    // Nested, so use from the embeddedMetaData
                    embMmd = ownerMmd.getEmbeddedMetaData().getMemberMetaData()[fieldNumber];
                }

                // TODO Cater for inherited embedded objects (discriminator)

                ObjectProvider embOP = ec.findObjectProviderForEmbedded(value, op, mmd);
                FieldManager ffm = new StoreEmbeddedFieldManager(embOP, node, insert, embMmd);
                embOP.provideFields(embcmd.getAllMemberPositions(), ffm);
                return;
            }
            // TODO Support embedded objects
            throw new NucleusUserException("Don't currently support embedded field : " + mmd.getFullFieldName());
        }

        if (mmd.isSerialized())
        {
            throw new NucleusUserException("Field " + mmd.getFullFieldName() + " is marked as serialised, but serialised is not yet supported");
        }

        if (Relation.isRelationSingleValued(relationType))
        {
            // 1-1/N-1 Make sure it is persisted and form the relation
            Object valuePC = (value != null ? ec.persistObjectInternal(value, null, -1, -1) : null);
            ObjectProvider relatedOP = (value != null ? ec.findObjectProvider(valuePC) : null);
            if (relationType != Relation.MANY_TO_ONE_BI && mmd.getMappedBy() == null)
            {
                // Only have a Relationship if this side owns the relation
                Node relatedNode = 
                    (value != null ? Neo4jUtils.getNodeForObjectProvider(node.getGraphDatabase(), relatedOP) : null);

                boolean hasRelation = false;
                if (!insert)
                {
                    // Check for old value and remove Relationship if to a different Node
                    Iterable<Relationship> rels = node.getRelationships(DNRelationshipType.SINGLE_VALUED);
                    Iterator<Relationship> relIter = rels.iterator();
                    while (relIter.hasNext())
                    {
                        Relationship rel = relIter.next();
                        if (rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME).equals(mmd.getName()))
                        {
                            // Check if existing relationship for this field is to the same node
                            Node currentNode = rel.getOtherNode(node);
                            if (currentNode.equals(relatedNode))
                            {
                                hasRelation = true;
                                break;
                            }
                            else
                            {
                                // Remove old Relationship TODO Cascade delete?
                                rel.delete();
                            }
                        }
                    }
                }

                if (!hasRelation && relatedNode != null)
                {
                    // Add the new Relationship
                    Relationship rel = node.createRelationshipTo(relatedNode, DNRelationshipType.SINGLE_VALUED);
                    rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME, mmd.getName());
                    if (Relation.isBidirectional(relationType))
                    {
                        AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                        rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME_NONOWNER, relMmds[0].getName());
                    }
                }
            }
            return;
        }
        else if (Relation.isRelationMultiValued(relationType))
        {
            if (mmd.hasCollection())
            {
                Collection coll = (Collection)value;
                List<Node> collNodes = null;
                if (relationType == Relation.ONE_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_UNI ||
                    (relationType == Relation.MANY_TO_MANY_BI && mmd.getMappedBy() == null))
                {
                    collNodes = new ArrayList<Node>();
                }

                if (value != null)
                {
                    // Persist any nodes
                    Iterator collIter = coll.iterator();
                    while (collIter.hasNext())
                    {
                        if (mmd.getCollection().isSerializedElement())
                        {
                            // TODO Support Serialised elements
                            throw new NucleusUserException("Don't currently support serialised collection elements at " +
                                mmd.getFullFieldName());
                        }

                        Object element = collIter.next();
                        if (element != null)
                        {
                            Object elementPC = ec.persistObjectInternal(element, null, -1, -1);
                            ObjectProvider relatedOP = ec.findObjectProvider(elementPC);
                            if (relationType == Relation.ONE_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_UNI ||
                                (relationType == Relation.MANY_TO_MANY_BI && mmd.getMappedBy() == null))
                            {
                                // Only add when this side owns the relation
                                Node relatedNode = Neo4jUtils.getNodeForObjectProvider(node.getGraphDatabase(), relatedOP);
                                collNodes.add(relatedNode);
                            }
                        }
                        else
                        {
                            throw new NucleusUserException("Dont currently support having null elements in collections : " + mmd.getFullFieldName());
                        }
                    }
                }
                if (relationType == Relation.ONE_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_UNI ||
                    (relationType == Relation.MANY_TO_MANY_BI && mmd.getMappedBy() == null))
                {
                    // Owner of the relation, so manage the Relationship objects
                    if (insert)
                    {
                        // Insert of the collection, so create Relationship for each
                        int index = 0;
                        for (Node newNode : collNodes)
                        {
                            Relationship rel = node.createRelationshipTo(newNode, DNRelationshipType.MULTI_VALUED);
                            rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME, mmd.getName());
                            if (coll instanceof List)
                            {
                                rel.setProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME, index);
                                index++;
                            }
                            if (Relation.isBidirectional(relationType))
                            {
                                AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                                rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME_NONOWNER, relMmds[0].getName());
                            }
                        }
                    }
                    else
                    {
                        // Update of the collection so remove existing Relationship and create new
                        Iterable<Relationship> rels = node.getRelationships(DNRelationshipType.MULTI_VALUED);
                        Iterator<Relationship> relIter = rels.iterator();
                        while (relIter.hasNext())
                        {
                            Relationship rel = relIter.next();
                            if (rel.getProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME).equals(mmd.getName()))
                            {
                                rel.delete();
                            }
                        }

                        int index = 0;
                        for (Node newNode : collNodes)
                        {
                            Relationship rel = node.createRelationshipTo(newNode, DNRelationshipType.MULTI_VALUED);
                            rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME, mmd.getName());
                            if (coll instanceof List)
                            {
                                rel.setProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME, index);
                                index++;
                            }
                            if (Relation.isBidirectional(relationType))
                            {
                                AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                                rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME_NONOWNER, relMmds[0].getName());
                            }
                        }
                    }
                }
            }
            else if (mmd.hasArray())
            {
                // TODO Doesn't cater for updates to the array and management of the Relationship objects
                for (int i=0;i<Array.getLength(value);i++)
                {
                    if (mmd.getArray().isSerializedElement())
                    {
                        // TODO Support Serialised elements
                        throw new NucleusUserException("Don't currently support serialised collection elements at " +
                            mmd.getFullFieldName());
                    }

                    Object element = Array.get(value, i);
                    if (element != null)
                    {
                        Object elementPC = ec.persistObjectInternal(element, null, -1, -1);
                        ObjectProvider relatedOP = ec.findObjectProvider(elementPC);
                        if (relationType == Relation.ONE_TO_MANY_BI || relationType == Relation.ONE_TO_MANY_UNI ||
                            (relationType == Relation.MANY_TO_MANY_BI && mmd.getMappedBy() == null))
                        {
                            // Only add when this side owns the relation
                            Node relatedNode = Neo4jUtils.getNodeForObjectProvider(node.getGraphDatabase(), relatedOP);
                            Relationship rel = node.createRelationshipTo(relatedNode, DNRelationshipType.MULTI_VALUED);
                            rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME, mmd.getName());
                            rel.setProperty(Neo4jStoreManager.RELATIONSHIP_INDEX_NAME, i);
                            if (Relation.isBidirectional(relationType))
                            {
                                AbstractMemberMetaData[] relMmds = mmd.getRelatedMemberMetaData(clr);
                                rel.setProperty(Neo4jStoreManager.RELATIONSHIP_FIELD_NAME_NONOWNER, relMmds[0].getName());
                            }
                        }
                    }
                    else
                    {
                        throw new NucleusUserException("Dont currently support having null elements in array : " + mmd.getFullFieldName());
                    }
                }
            }
            else if (mmd.hasMap())
            {
                // TODO Persist 1-N
                throw new NucleusUserException("Don't currently support 1-N/M-N maps : " + mmd.getFullFieldName());
            }
            return;
        }

        if (value == null)
        {
            // Don't add the property when null TODO This would likely cause problems with queries!
            // Could store a special String value against it maybe
            return;
        }

        if (mmd.getTypeConverterName() != null)
        {
            // User-defined type converter
            TypeManager typeMgr = op.getExecutionContext().getNucleusContext().getTypeManager();
            TypeConverter conv = typeMgr.getTypeConverterForName(mmd.getTypeConverterName());
            node.setProperty(propName, conv.toDatastoreType(value));
        }
        else
        {
            Object storedValue = Neo4jUtils.getStoredValueForField(ec, mmd, value, FieldRole.ROLE_FIELD);
            node.setProperty(propName, storedValue);
        }
        op.wrapSCOField(fieldNumber, value, false, false, true);
    }
}